Package: fastVoteR
Title: Efficient Voting Methods for Committee Selection
Version: 0.0.3
Authors@R: c(
    person("John", "Zobolas", , "bblodfon@gmail.com", role = c("cre", "aut"),
           comment = c(ORCID = "0000-0002-3609-8674")),
    person("Anne-Marie", "George", role = "ctb",
           comment = c(ORCID = "0000-0001-9232-8211"))
  )
Description: A fast 'Rcpp'-based implementation of polynomially-computable
    voting theory methods for committee ranking and scoring. The package
    includes methods such as Approval Voting (AV), Satisfaction Approval
    Voting (SAV), sequential Proportional Approval Voting (PAV), and
    sequential Phragmen's Rule. Weighted variants of these methods are
    also provided, allowing for differential voter influence.
License: MIT + file LICENSE
URL: https://bblodfon.github.io/fastVoteR/
Imports: checkmate, Rcpp
Suggests: testthat (>= 3.0.0)
LinkingTo: Rcpp
Config/testthat/edition: 3
Encoding: UTF-8
RoxygenNote: 7.3.3
NeedsCompilation: yes
Packaged: 2026-05-15 15:17:20 UTC; john
Author: John Zobolas [cre, aut] (ORCID:
    <https://orcid.org/0000-0002-3609-8674>),
  Anne-Marie George [ctb] (ORCID:
    <https://orcid.org/0000-0001-9232-8211>)
Maintainer: John Zobolas <bblodfon@gmail.com>
Repository: CRAN
Date/Publication: 2026-05-15 20:20:13 UTC
Built: R 4.5.2; x86_64-apple-darwin20; 2026-05-15 21:55:53 UTC; unix
Archs: fastVoteR.so.dSYM

Package: fastVoteR
Title: Efficient Voting Methods for Committee Selection
Version: 0.0.3
Authors@R: c(
    person("John", "Zobolas", , "bblodfon@gmail.com", role = c("cre", "aut"),
           comment = c(ORCID = "0000-0002-3609-8674")),
    person("Anne-Marie", "George", role = "ctb",
           comment = c(ORCID = "0000-0001-9232-8211"))
  )
Description: A fast 'Rcpp'-based implementation of polynomially-computable
    voting theory methods for committee ranking and scoring. The package
    includes methods such as Approval Voting (AV), Satisfaction Approval
    Voting (SAV), sequential Proportional Approval Voting (PAV), and
    sequential Phragmen's Rule. Weighted variants of these methods are
    also provided, allowing for differential voter influence.
License: MIT + file LICENSE
URL: https://bblodfon.github.io/fastVoteR/
Imports: checkmate, Rcpp
Suggests: testthat (>= 3.0.0)
LinkingTo: Rcpp
Config/testthat/edition: 3
Encoding: UTF-8
RoxygenNote: 7.3.3
NeedsCompilation: yes
Packaged: 2026-05-15 15:17:20 UTC; john
Author: John Zobolas [cre, aut] (ORCID:
    <https://orcid.org/0000-0002-3609-8674>),
  Anne-Marie George [ctb] (ORCID:
    <https://orcid.org/0000-0001-9232-8211>)
Maintainer: John Zobolas <bblodfon@gmail.com>
Repository: CRAN
Date/Publication: 2026-05-15 20:20:13 UTC
Built: R 4.6.0; x86_64-apple-darwin20; 2026-05-15 22:32:08 UTC; unix
Archs: fastVoteR.so.dSYM

# Generated by roxygen2: do not edit by hand

export(av)
export(rank_candidates)
export(sav)
export(seq_pav)
export(seq_phragmen)
import(checkmate)
importFrom(Rcpp,sourceCpp)
useDynLib(fastVoteR, .registration = TRUE)

# fastVoteR

[Package website](https://bblodfon.github.io/fastVoteR/)

<!-- badges: start -->

[![R-CMD-check](https://github.com/bblodfon/fastVoteR/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/bblodfon/fastVoteR/actions/workflows/R-CMD-check.yaml)
[![Codecov test
coverage](https://codecov.io/gh/bblodfon/fastVoteR/graph/badge.svg)](https://app.codecov.io/gh/bblodfon/fastVoteR)
[![CRAN
status](https://www.r-pkg.org/badges/version/fastVoteR)](https://CRAN.R-project.org/package=fastVoteR)
<!-- badges: end -->

`fastVoteR` is an R package with efficient voting methods for committee
selection. Uses [Rcpp](https://cran.r-project.org/package=Rcpp).

## Installation

CRAN version:

``` r
install.packages("fastVoteR")
```

Development version from GitHub:

``` r
# install.packages("pak")
pak::pak("bblodfon/fastVoteR")
```

## Example

``` r
library(fastVoteR)

# 5 candidates
candidates = paste0("V", seq_len(5))
candidates
#> [1] "V1" "V2" "V3" "V4" "V5"

# 4 voters with their preferences (approval ballots)
voters = list(
  c("V3", "V1", "V4"),
  c("V3", "V1"),
  c("V3", "V2"),
  c("V2", "V4")
)

set.seed(42)

# voter weights
weights = c(1.1, 2.5, 0.8, 0.9)

# Approval voting (all voters equal)
rank_candidates(voters, candidates)
#>    candidate score norm_score borda_score
#>       <char> <num>      <num>       <num>
#> 1:        V3     3       0.75        1.00
#> 2:        V1     2       0.50        0.75
#> 3:        V2     2       0.50        0.50
#> 4:        V4     2       0.50        0.25
#> 5:        V5     0       0.00        0.00

# Approval voting (voters unequal)
rank_candidates(voters, candidates, weights)
#>    candidate score norm_score borda_score
#>       <char> <num>      <num>       <num>
#> 1:        V3   4.4  0.8301887        1.00
#> 2:        V1   3.6  0.6792453        0.75
#> 3:        V4   2.0  0.3773585        0.50
#> 4:        V2   1.7  0.3207547        0.25
#> 5:        V5   0.0  0.0000000        0.00

# Satisfaction Approval voting (voters unequal)
rank_candidates(voters, candidates, weights, method = "sav")
#>    candidate     score norm_score borda_score
#>       <char>     <num>      <num>       <num>
#> 1:        V3 2.0166667  0.8175676        1.00
#> 2:        V1 1.6166667  0.6554054        0.75
#> 3:        V2 0.8500000  0.3445946        0.50
#> 4:        V4 0.8166667  0.3310811        0.25
#> 5:        V5 0.0000000  0.0000000        0.00

# Sequential Proportional Approval voting (voters equal)
rank_candidates(voters, candidates, method = "seq_pav")
#>    candidate borda_score
#>       <char>       <num>
#> 1:        V3        1.00
#> 2:        V2        0.75
#> 3:        V1        0.50
#> 4:        V4        0.25
#> 5:        V5        0.00
```

## Related work

See [vote](https://CRAN.R-project.org/package=vote) and
[votesys](https://CRAN.R-project.org/package=votesys) R packages. For
strictly ABC-voting rules, see
[abcvoting](https://github.com/martinlackner/abcvoting) Python package.

## Code of Conduct

Please note that the `fastVoteR` project is released with a [Contributor
Code of
Conduct](https://bblodfon.github.io/fastVoteR/CODE_OF_CONDUCT.html). By
contributing to this project, you agree to abide by its terms.

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

av_rcpp <- function(voters, candidates, weights) {
    .Call(`_fastVoteR_av_rcpp`, voters, candidates, weights)
}

seq_pav_rcpp <- function(voters, candidates, weights, committee_size) {
    .Call(`_fastVoteR_seq_pav_rcpp`, voters, candidates, weights, committee_size)
}

seq_phragmen_rcpp <- function(voters, candidates, weights, committee_size) {
    .Call(`_fastVoteR_seq_phragmen_rcpp`, voters, candidates, weights, committee_size)
}

sav_rcpp <- function(voters, candidates, weights) {
    .Call(`_fastVoteR_sav_rcpp`, voters, candidates, weights)
}

# All the following R methods are wrappers for the Rcpp implementations

#' @title Approval Voting
#'
#' @description
#' **Approval Voting** (AV) ranks candidates based on the number of voters approving them.
#'
#' This function uses an internal C++ implementation for efficient computation.
#' For equal weights, a faster R implementation is used.
#'
#' @template param_voters
#' @template param_candidates
#' @template param_weights
#' @template param_committee_size
#' @template param_borda_score
#' @template param_check
#'
#' @return A `data.frame` with columns:
#' - `"candidate"`: Candidate names.
#' - `"score"`: Approval scores.
#' - `"norm_score"`: Normalized scores, scaled to the range \eqn{[0,1]}.
#' - `"borda_score"`: Borda scores for method-agnostic comparison, ranging in
#' \eqn{[0,1]}, where the top candidate receives a score of 1 and the lowest-ranked
#' candidate receives a score of 0, based on the total number of candidates.
#'
#' Candidates are ordered by decreasing `"score"`.
#' @family voting methods
#' @export
av = function(
  voters,
  candidates,
  weights = NULL,
  committee_size = NULL,
  borda_score = TRUE,
  check = FALSE
) {
  assert_inputs(voters, candidates, weights, committee_size, borda_score, check)

  if (is.null(weights)) {
    # Assign equal weights to all voters
    weights = rep(1, length(voters))
  }

  if (check) {
    check_voters(voters, candidates)
  }

  # faster R version in case of equal weights
  if (all(weights == 1)) {
    count_tbl = sort(table(unlist(voters)), decreasing = TRUE)
    candidates_selected = names(count_tbl)
    candidates_not_selected = setdiff(candidates, candidates_selected)
    approval_counts = as.vector(count_tbl)

    res_sel = data.frame(
      candidate = candidates_selected,
      score = approval_counts,
      norm_score = approval_counts / length(voters),
      stringsAsFactors = FALSE
    )

    # candidates not selected at all get a score of 0
    res_not_sel = if (length(candidates_not_selected) == 0) {
      data.frame(
        candidate = character(0),
        score = numeric(0),
        norm_score = numeric(0),
        stringsAsFactors = FALSE
      )
    } else {
      data.frame(
        candidate = candidates_not_selected,
        score = 0,
        norm_score = 0,
        stringsAsFactors = FALSE
      )
    }

    res = rbind(res_sel, res_not_sel)
  } else {
    # returns AV scores so needs ordering
    res = data.frame(av_rcpp(voters, candidates, weights), stringsAsFactors = FALSE)
    res = res[order(res$score, decreasing = TRUE), ]
    rownames(res) = NULL
  }

  # subset to top N rows
  if (!is.null(committee_size)) res = res[1:committee_size, ]

  # add borda scores
  if (borda_score) add_borda_score(res, n = length(candidates)) else res
}

#' @title Satisfaction Approval Voting
#'
#' @description
#' **Satisfaction Approval Voting** (SAV) ranks candidates by normalizing
#' approval scores based on the size of each voter's approval set. Voters who
#' approve more candidates contribute a lesser score to the individual approved
#' candidates.
#'
#' This function uses an internal C++ implementation for efficient computation.
#'
#' @template param_voters
#' @template param_candidates
#' @template param_weights
#' @template param_committee_size
#' @template param_borda_score
#' @template param_check
#'
#' @return A `data.frame` with columns:
#' - `"candidate"`: Candidate names.
#' - `"score"`: Satisfaction scores.
#' - `"norm_score"`: Normalized scores, scaled to the range \eqn{[0,1]}.
#' - `"borda_score"`: Borda scores for method-agnostic comparison, ranging in
#' \eqn{[0,1]}, where the top candidate receives a score of 1 and the lowest-ranked
#' candidate receives a score of 0, based on the total number of candidates.
#'
#' Candidates are ordered by decreasing `"score"`.
#' @family voting methods
#' @export
sav = function(
  voters,
  candidates,
  weights = NULL,
  committee_size = NULL,
  borda_score = TRUE,
  check = FALSE
) {
  assert_inputs(voters, candidates, weights, committee_size, borda_score, check)

  if (is.null(weights)) {
    # Assign equal weights to all voters
    weights = rep(1, length(voters))
  }

  if (check) {
    check_voters(voters, candidates)
  }

  # returns SAV scores so needs ordering
  res = data.frame(sav_rcpp(voters, candidates, weights), stringsAsFactors = FALSE)
  res = res[order(res$score, decreasing = TRUE), ]
  rownames(res) = NULL

  # subset to top N rows
  if (!is.null(committee_size)) res = res[1:committee_size, ]

  # add borda scores
  if (borda_score) add_borda_score(res, n = length(candidates)) else res
}

#' @title Sequential Proportional Approval Voting
#'
#' @description
#' **Sequential Proportional Approval Voting** (SeqPAV) is a multi-winner method
#' that builds a committee by iteratively maximizing a proportionality-based
#' satisfaction score. After each selection, the weights of voters who approved
#' the chosen candidate are reduced, which promotes proportional representation.
#' The **PAV score** is computed as the weighted sum of harmonic numbers based on
#' how many elected candidates each voter supports. The process continues until
#' the specified committee size is reached or all candidates are ranked.
#'
#' This function uses an internal C++ implementation for efficient computation.
#'
#' @template param_voters
#' @template param_candidates
#' @template param_weights
#' @template param_committee_size
#' @template param_borda_score
#' @template param_check
#'
#' @return A `data.frame` with columns:
#' - `"candidate"`: Candidate names.
#' - `"borda_score"`: Borda scores for method-agnostic comparison, ranging in \eqn{[0,1]}, where the top candidate receives a score of 1 and the lowest-ranked candidate receives a score of 0, based on the total number of candidates.
#'
#' Candidates are ordered by the sequence in which they were selected.
#' @family voting methods
#' @export
seq_pav = function(
  voters,
  candidates,
  weights = NULL,
  committee_size = NULL,
  borda_score = TRUE,
  check = FALSE
) {
  assert_inputs(voters, candidates, weights, committee_size, borda_score, check)

  if (is.null(weights)) {
    # Assign equal weights to all voters
    weights = rep(1, length(voters))
  }

  if (check) {
    check_voters(voters, candidates)
  }

  # set committee_size to total number of candidates if not specified
  if (is.null(committee_size)) committee_size = length(candidates)

  # returns ranked candidates from best to worst (up to committee_size)
  res = data.frame(seq_pav_rcpp(voters, candidates, weights, committee_size),
                   stringsAsFactors = FALSE)

  # add borda scores
  if (borda_score) add_borda_score(res, n = length(candidates)) else res
}

#' @title Sequential Phragmen's Rule
#'
#' @description
#' **Sequential Phragmen's Rule** is a multi-winner method that builds a committee
#' by distributing representation *loads* across voters as evenly as possible.
#' At each step, it selects the candidate that yields the smallest increase in
#' voter load, then updates loads for voters who approved the chosen candidate.
#' The process continues until the committee is filled or all candidates are ranked.
#'
#' This function uses an internal C++ implementation for efficient computation.
#'
#' @template param_voters
#' @template param_candidates
#' @template param_weights
#' @template param_committee_size
#' @template param_borda_score
#' @template param_check
#'
#' @return A `data.frame` with columns:
#' - `"candidate"`: Candidate names.
#' - `"borda_score"`: Borda scores for method-agnostic comparison, ranging in \eqn{[0,1]}, where the top candidate receives a score of 1 and the lowest-ranked candidate receives a score of 0, based on the total number of candidates.
#'
#' Candidates are ordered by the sequence in which they were selected.
#' @family voting methods
#' @export
seq_phragmen = function(
  voters,
  candidates,
  weights = NULL,
  committee_size = NULL,
  borda_score = TRUE,
  check = FALSE
) {
  assert_inputs(voters, candidates, weights, committee_size, borda_score, check)

  if (is.null(weights)) {
    # Assign equal weights to all voters
    weights = rep(1, length(voters))
  }

  if (check) {
    check_voters(voters, candidates)
  }

  # set committee_size to total number of candidates if not specified
  if (is.null(committee_size)) committee_size = length(candidates)

  # returns ranked candidates from best to worst (up to committee_size)
  res = data.frame(seq_phragmen_rcpp(voters, candidates, weights, committee_size))

  # add borda scores
  if (borda_score) add_borda_score(res, n = length(candidates)) else res
}

#' @title Adds normalized borda scores
#' @note `n` needs to be the total number of candidates (irrespective of committee size)
#' @noRd
#' @keywords internal
add_borda_score = function(df, n) {
  assert_number(n, null.ok = FALSE, lower = 1)

  n_rows = nrow(df)
  df$borda_score = if (n_rows == 1) 1 else (n - seq_len(n_rows)) / (n - 1)

  df
}

#' @title Check voter approvals
#' @description
#' Validates the approval ballots in `voters`.
#' Ensures:
#' - Each voter approves at least one candidate.
#' - No voter lists the same candidate more than once.
#' - Every approved candidate appears in `candidates`.
#' @noRd
#' @keywords internal
check_voters = function(voters, candidates) {
  voter_lengths = lengths(voters)

  if (any(voter_lengths == 0L)) {
    stop("all voters must approve at least one candidate", call. = FALSE)
  }

  voted_candidates = unique(unlist(voters, use.names = FALSE))
  candidate_matches = match(voted_candidates, candidates, nomatch = NA_integer_)
  if (anyNA(candidate_matches)) {
    stop("all voted candidates must be present in the 'candidates' vector", call. = FALSE)
  }

  if (any(vapply(voters, anyDuplicated, integer(1)) > 0L)) {
    stop("voters must not contain duplicated candidates", call. = FALSE)
  }
}

#' @noRd
#' @keywords internal
assert_inputs = function(voters, candidates, weights, committee_size, borda_score, check) {
  # input checks
  assert_list(voters, min.len = 1, types = "character", null.ok = FALSE, any.missing = FALSE)
  assert_character(candidates, min.len = 1, null.ok = FALSE, any.missing = FALSE)
  # each voter's weight must be non-negative
  assert_numeric(weights, len = length(voters), lower = 0, any.missing = FALSE, null.ok = TRUE)
  # at least one positive weight is required (if weights are provided)
  if (!is.null(weights)) assert_true(sum(weights) > 0)
  assert_int(committee_size, lower = 1, null.ok = TRUE)
  assert_flag(borda_score)
  assert_flag(check)
}

# see `helper.R` for example inputs (`vot2`, `cand2`, `w2` ) and `test_*` functions

test_that("satisfaction approval voting", {
  test_large_example(sav)
  test_borda_score(sav)
  test_equal_votes(sav)

  # small data example
  res = sav(vot2, cand2, w2) # different weights
  expect_equal(res$candidate, c("V3", "V1", "V4", "V2", "V5"))
  expect_equal(res[candidate == "V3", norm_score], 1) # voted by all
  expect_equal(res[candidate == "V5", norm_score], 0) # never voted

  res2 = sav(vot2, cand2, we2) # equal weights
  expect_equal(res2$candidate, c("V3", "V1", "V4", "V2", "V5"))
})

# see `helper.R` for example inputs (`vot2`, `cand2`, `w2` ) and `test_*` functions

test_that("sequential Phragmen's rule", {
  test_large_example(seq_phragmen, committee_size = 5)
  test_borda_score(seq_phragmen)
  test_equal_votes(seq_phragmen)

  # small data example
  res = seq_phragmen(vot2, cand2, w2) # different weights
  expect_equal(res$candidate, c("V3", "V1", "V4", "V2", "V5"))

  res2 = sav(vot2, cand2, we2) # equal weights
  expect_equal(res2$candidate, c("V3", "V1", "V4", "V2", "V5"))

  # Example 2.9 from Lackner's "Multi-Winner Voting with Approval Preferences"
  cand3 = paste0("V", 1:7)
  vot3 = list(
    c("V1", "V2"),
    c("V1", "V2"),
    c("V1", "V2"),
    c("V1", "V3"),
    c("V1", "V3"),
    c("V1", "V3"),
    c("V1", "V4"),
    c("V1", "V4"),
    c("V2", "V3", "V6"),
    "V5", "V6", "V7"
  )
  we3 = rep(1, length(vot3)) # equal weights

  # output committee should be:
  # - V1
  # - {V2, V3} => tied
  # - V4
  # - V6
  # - {V5, V7} => tied
  #
  # Rankings after resampling the candidates should respect the above ordering
  res = lapply(1:10, function(i) {
    sp = seq_phragmen(vot3, sample(cand3), we3)
    sp$candidate
  })

  for (i in 1:10) {
    committee = res[[i]]
    expect_equal(committee[1], "V1")
    expect_contains(committee[2:3], c("V2", "V3"))
    expect_equal(committee[4:5], c("V4", "V6"))
    expect_contains(committee[6:7], c("V5", "V7"))
  }
})

# see `helper.R` for example inputs (`vot2`, `cand2`, `w2` ) and `test_*` functions

test_that("sequential proportional approval voting", {
  test_large_example(seq_pav, committee_size = 5)
  test_borda_score(seq_pav)
  test_equal_votes(seq_pav)

  # small data example
  res = seq_pav(vot2, cand2, w2) # different weights
  expect_equal(res$candidate, c("V3", "V1", "V4", "V2", "V5"))

  res2 = seq_pav(vot2, cand2, we2) # equal weights
  expect_equal(res2$candidate, c("V3", "V1", "V4", "V2", "V5"))
})

#' @title Rank candidates based on voter preferences
#'
#' @description
#' Calculates a ranking of candidates based on voters' preferences.
#' Approval-Based Committe (ABC) rules are based on Lackner et al. (2023).
#' For an example use in Machine Learning for ensemble feature selection,
#' see Zobolas et al. (2026).
#'
#' @details
#' We implement several consensus-based ranking methods, where voters express
#' preferences for candidates.
#' The framework has three components:
#' - **Voters**: A list where each element is the set of approved candidates for
#'   a single voter.
#' - **Candidates**: A character vector of all possible candidates. This vector
#'   can be shuffled to randomize tie-breaking across methods.
#' - **Weights**: A numeric vector giving each voter’s influence. Equal weights
#'   mean equal influence; differing weights reflect varying importance.
#'
#' This function is a thin wrapper that dispatches to method-specific implementations.
#' Supported methods include:
#'
#' 1. Approval Voting (`method = "av"`), calls [av()]
#' 2. Satisfaction Approval Voting (`method = "sav"`), calls [sav()]
#' 3. Sequential Proportional Approval Voting (`method = "seq_pav"`), calls [seq_pav()]
#' 4. Sequential Phragmen’s Rule (`method = "seq_phragmen"`), calls [seq_phragmen()]
#'
#' All methods support voter weights.
#'
#' For method-agnostic comparisons, we compute **borda scores**, which map each
#' candidate’s rank to a normalized scale across methods that may otherwise
#' return different score scales or only ordinal rankings.
#'
#' For sequential methods such as `"seq_pav"` and `"seq_phragmen"`, the
#' `committee_size` parameter can speed up computation by selecting only the top
#' \eqn{N} candidates instead of producing a full ranking. For non-sequential methods
#' (e.g., `"sav"` or `"av"`), it simply truncates the final ranking to the top
#' \eqn{N} candidates.
#'
#' @template param_voters
#' @template param_candidates
#' @template param_weights
#' @template param_committee_size
#' @template param_borda_score
#' @template param_check
#' @param method (`character(1)`) \cr
#'  The ranking voting method to use. Must be one of: `"av"`, `"sav"`, `"seq_pav"`,
#'  `"seq_phragmen"`. See Details.
#' @param shuffle_candidates (`logical(1)`) \cr
#'  Whether to randomly shuffle candidates before ranking.
#'  This provides random tie-breaking and avoids deterministic bias when scores
#'  are equal. Default is `TRUE`.
#'
#' @return A `data.frame` with columns:
#' - `"candidate"`: Candidate names.
#' - `"score"`: Scores assigned to each candidate based on the selected method
#' (if applicable).
#' - `"norm_score"`: Normalized scores (if applicable), scaled to the range
#' \eqn{[0,1]}, which can be loosely interpreted as **selection probabilities**
#' (see Meinshausen et al. (2010) for an example in Machine Learning research
#' where the goal is to perform stable feature selection).
#' - `"borda_score"`: Borda scores for method-agnostic comparison, ranging in
#' \eqn{[0,1]}, where the top candidate receives a score of 1 and the lowest-ranked
#' candidate receives a score of 0.
#'
#' Candidates are ordered by decreasing `"score"`, or by `"borda_score"` if the
#' method returns only rankings.
#'
#' @references
#' Lackner M, Skowron P (2023). *Multi-Winner Voting with Approval Preferences*.
#' Springer Nature, 121 p. \doi{10.1007/978-3-031-09016-5}.
#'
#' Zobolas J, George A, Lopez A, Fischer S, Becker M, Aittokallio T (2026).
#' "Prognostic biomarker discovery in pancreatic cancer through hybrid ensemble
#' feature selection and multi-omics data." *BioData Mining*. \doi{10.1186/s13040-026-00546-0}.
#'
#' Meinshausen N, Buhlmann P (2010). "Stability Selection." *Journal of the Royal
#' Statistical Society Series B: Statistical Methodology*, 72(4), 417-473.
#' \doi{10.1111/J.1467-9868.2010.00740.X}.
#'
#' @examples
#' # 5 candidates
#' candidates = paste0("V", seq_len(5))
#'
#' # 4 voters
#' voters = list(
#'   c("V3", "V1", "V4"),
#'   c("V3", "V1"),
#'   c("V3", "V2"),
#'   c("V2", "V4")
#' )
#'
#' # voter weights
#' weights = c(1.1, 2.5, 0.8, 0.9)
#'
#' # Approval voting (all voters equal)
#' rank_candidates(voters, candidates)
#'
#' # Approval voting (voters unequal)
#' rank_candidates(voters, candidates, weights)
#'
#' # Satisfaction Approval voting (voters unequal, no borda score)
#' rank_candidates(voters, candidates, weights, method = "sav", borda_score = FALSE)
#'
#' # Sequential Proportional Approval voting (voters equal, no borda score)
#' rank_candidates(voters, candidates, method = "seq_pav", borda_score = FALSE)
#'
#' # Sequential Phragmen's Rule (voters equal)
#' rank_candidates(voters, candidates, method = "seq_phragmen", borda_score = FALSE)
#'
#' @export
rank_candidates = function(
    voters,
    candidates,
    weights = NULL,
    committee_size = NULL,
    method = "av",
    borda_score = TRUE,
    shuffle_candidates = TRUE,
    check = FALSE)
  {
  assert_choice(method, choices = c("av", "sav", "seq_pav", "seq_phragmen"))
  assert_flag(shuffle_candidates)

  # Shuffle candidates for consistent tie-breaking
  if (shuffle_candidates) {
    candidates = sample(candidates)
  }

  # Call the appropriate ranking method
  if (method == "av") {
    res = av(voters, candidates, weights, committee_size, borda_score, check)
  } else if (method == "sav") {
    res = sav(voters, candidates, weights, committee_size, borda_score, check)
  } else if (method == "seq_pav") {
    res = seq_pav(voters, candidates, weights, committee_size, borda_score, check)
  } else if (method == "seq_phragmen") {
    res = seq_phragmen(voters, candidates, weights, committee_size, borda_score, check)
  }

  res
}

# see `helper.R` for example inputs (`vot2`, `cand2`, `w2` )

test_that("rank candidates", {
  # invalid method
  expect_error(rank_candidates(voters = vot2, candidates = cand2, method = "XYZ"))
  # invalid shuffle_candidates
  expect_error(rank_candidates(voters = vot2, candidates = cand2, shuffle_candidates = NULL))

  # verify each method returns a data frame with expected dimensions (equal weights)
  expect_data_frame(rank_candidates(vot2, cand2, method = "av"),
                    nrows = length(cand2), ncols = 4, any.missing = FALSE)
  expect_data_frame(rank_candidates(vot2, cand2, method = "sav"),
                    nrows = length(cand2), ncols = 4, any.missing = FALSE)
  expect_data_frame(rank_candidates(vot2, cand2, method = "seq_pav"),
                    nrows = length(cand2), ncols = 2, any.missing = FALSE)
  expect_data_frame(rank_candidates(vot2, cand2, method = "seq_phragmen"),
                    nrows = length(cand2), ncols = 2, any.missing = FALSE)

  # verify each method returns a data frame with expected dimensions (unequal weights)
  expect_data_frame(rank_candidates(vot2, cand2, w2, method = "av"),
                    nrows = length(cand2), ncols = 4, any.missing = FALSE)
  expect_data_frame(rank_candidates(vot2, cand2, w2, method = "sav"),
                    nrows = length(cand2), ncols = 4, any.missing = FALSE)
  expect_data_frame(rank_candidates(vot2, cand2, w2, method = "seq_pav"),
                    nrows = length(cand2), ncols = 2, any.missing = FALSE)
  expect_data_frame(rank_candidates(vot2, cand2, w2, method = "seq_phragmen"),
                    nrows = length(cand2), ncols = 2, any.missing = FALSE)
})

# see `helper.R` for example inputs (`vot2`, `cand2`, `w2` ) and `test_*` functions

test_that("approval voting", {
  test_large_example(av)
  test_borda_score(av)
  test_equal_votes(av)

  # small data example
  res = av(vot2, cand2, w2) # different weights
  expect_equal(res$candidate, c("V3", "V1", "V4", "V2", "V5"))
  expect_equal(res[candidate == "V3", norm_score], 1) # voted by all
  expect_equal(res[candidate == "V5", norm_score], 0) # never voted

  res2 = av(vot2, cand2, we2) # equal weights
  expect_equal(res2$candidate[1:2], c("V3", "V1"))
  # due to equal scores better check the following:
  expect_set_equal(res2$candidate[3:4], c("V2", "V4"))
  expect_equal(res2$candidate[5], "V5")
  expect_equal(res2$score, c(4, 2, 1, 1, 0)) # approval counts
  expect_equal(res2$norm_score, c(1, 0.5, 0.25, 0.25, 0))
})

# Create some example inputs for testing

# large data example
set.seed(42)

n_candidates = 800
n_voters = 200

cand = paste0("V", seq_len(n_candidates))
vot = lapply(1:n_voters, function(x) sample(cand, size = sample(2:30, 1)))
w = runif(n_voters)
we = rep(1, n_voters) # equal weights

# small data example
cand2 = paste0("V", seq_len(5))
# "V3" candidate in all sets, "V1" in half, "V2", "V4" once, "V5" nowhere!
vot2 = list(
  c("V3", "V1", "V2"),
  c("V3", "V1"),
  c("V3", "V4"),
  c("V3")
)
w2 = c(0.27, 0.23, 0.4, 0.5)
we2 = rep(1, length(vot2))

# edge case: all voters voted the same candidates, "V4" and "V5" are nowhere!
# equal/same votes
vot_equal = list(
  c("V3", "V1", "V2"),
  c("V3", "V1", "V2"),
  c("V3", "V1", "V2"),
  c("V3", "V1", "V2")
)

test_large_example = function(method_fun, committee_size = NULL) {
  # unequal weights
  res = method_fun(vot, cand, w, committee_size)
  size = ifelse(is.null(committee_size), length(cand), committee_size)

  expect_data_table(res, nrows = size, min.cols = 2, max.cols = 4)
  expect_contains(colnames(res), c("candidate", "borda_score"))

  # committee candidates are included in the candidates
  if (is.null(committee_size)) {
    expect_set_equal(res$candidate, cand)
  } else {
    expect_in(res$candidate, cand)
  }

  # scores must be positive
  if (!is.null(res$score)) {
    expect_true(all(res$score >= 0))
  }

  # normalized scores behave like probabilities
  if (!is.null(res$norm_score)) {
    expect_true(all(res$norm_score >= 0 & res$norm_score <= 1))
  }

  # house monotonicy check: ordering remains the same with less committee members
  res_less = method_fun(vot, cand, w, committee_size = 3)
  expect_equal(res_less$candidate, res$candidate[1:3])

  # equal weights
  res_equal = method_fun(vot, cand, we, committee_size)

  expect_data_table(res_equal, nrows = size, min.cols = 2, max.cols = 4)
  expect_contains(colnames(res_equal), c("candidate", "borda_score"))

  # committee candidates are included in the candidates
  if (is.null(committee_size)) {
    expect_set_equal(res_equal$candidate, cand)
  } else {
    expect_in(res_equal$candidate, cand)
  }

  # scores must be positive
  if (!is.null(res_equal$score)) {
    expect_true(all(res_equal$score >= 0))
  }

  # normalized scores behave like probabilities
  if (!is.null(res_equal$norm_score)) {
    expect_true(all(res_equal$norm_score >= 0 & res_equal$norm_score <= 1))
  }

  # house monotonicy check: ordering remains the same with less committee members
  res_equal_less = method_fun(vot, cand, we, committee_size = 3)
  expect_equal(res_equal_less$candidate, res_equal$candidate[1:3])

  # using different weights, candidate rankings should be different
  expect_false(identical(res$candidate, res_equal$candidate))
}

test_borda_score = function(method_fun) {
  # uses small data example
  # no borda score in the output result
  res = method_fun(vot2, cand2, w2, borda_score = FALSE)
  expect_error(expect_in("borda_score", colnames(res)))

  # borda score in the output result
  res = method_fun(vot2, cand2, w2, borda_score = TRUE)
  expect_in("borda_score", colnames(res))
  # borda scores make sense
  expect_equal(res$borda_score[1], 1) # top candidate
  expect_equal(res$borda_score[length(cand2)], 0) # lowest-ranked candidate

  # borda score takes into account the total number of candidates, so last score is not 0
  # when committee_size < #candidates
  res = method_fun(vot2, cand2, w2, committee_size = 2, borda_score = TRUE)
  expect_gt(res$borda_score[2] , 0)
}

test_equal_votes = function(method_fun) {
  res = method_fun(vot_equal, cand2, w2)
  expect_contains(res$candidate[1:3], c("V1", "V2", "V3"))
  expect_contains(res$candidate[4:5], c("V4", "V5")) # V4, V5 were never voted

  # if scores are present, V1, V2 and V3 should have the same positive scores,
  # V4 and V5 zero
  if (!is.null(res$score)) {
    expect_true(length(unique(res$score[1:3])) == 1) # all scores the same
    expect_true(all(res$score[1:3] > 0))
    expect_equal(res$score[4:5], c(0, 0))
  }
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/voting_methods.R
\name{av}
\alias{av}
\title{Approval Voting}
\usage{
av(
  voters,
  candidates,
  weights = NULL,
  committee_size = NULL,
  borda_score = TRUE,
  check = FALSE
)
}
\arguments{
\item{voters}{(\code{list()}) \cr
A list of subsets (\code{character} vectors), where each subset contains the
candidates approved or selected by a voter.}

\item{candidates}{(\code{character()}) \cr
A vector of all candidates to be ranked.}

\item{weights}{(\code{numeric()|NULL}) \cr
A numeric vector of non-negative weights representing each voter's influence.
Larger weight, higher influence.
Must have the same length as \code{voters}.
If \code{NULL} (default), all voters are assigned equal weights of 1, representing
equal influence.}

\item{committee_size}{(\code{integer(1)|NULL})\cr
Number of top-ranked candidates to return.
Default (\code{NULL}) returns all candidates.}

\item{borda_score}{(\code{logical(1)}) \cr
Whether to include a \code{borda_score} column in the output, which provides a
normalized score based on the candidate's rank.
If \code{TRUE} (default), the \code{borda_score} is calculated as \eqn{(p - i) / (p - 1)},
where \eqn{p} is the total number of candidates and \eqn{i} is the candidate's rank.}

\item{check}{(\code{logical(1)}) \cr
Whether to run additional voter-integrity checks. When \code{TRUE}, each voter
must approve at least one candidate, approvals must be unique per voter, and
all approved candidates must appear in \code{candidates}. Use \code{FALSE} to skip
these checks when inputs are known to be valid.}
}
\value{
A \code{data.frame} with columns:
\itemize{
\item \code{"candidate"}: Candidate names.
\item \code{"score"}: Approval scores.
\item \code{"norm_score"}: Normalized scores, scaled to the range \eqn{[0,1]}.
\item \code{"borda_score"}: Borda scores for method-agnostic comparison, ranging in
\eqn{[0,1]}, where the top candidate receives a score of 1 and the lowest-ranked
candidate receives a score of 0, based on the total number of candidates.
}

Candidates are ordered by decreasing \code{"score"}.
}
\description{
\strong{Approval Voting} (AV) ranks candidates based on the number of voters approving them.

This function uses an internal C++ implementation for efficient computation.
For equal weights, a faster R implementation is used.
}
\seealso{
Other voting methods: 
\code{\link{sav}()},
\code{\link{seq_pav}()},
\code{\link{seq_phragmen}()}
}
\concept{voting methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fastVoteR-package.R
\docType{package}
\name{fastVoteR-package}
\alias{fastVoteR}
\alias{fastVoteR-package}
\title{fastVoteR: Efficient Voting Methods for Committee Selection}
\description{
A fast 'Rcpp'-based implementation of polynomially-computable voting theory methods for committee ranking and scoring. The package includes methods such as Approval Voting (AV), Satisfaction Approval Voting (SAV), sequential Proportional Approval Voting (PAV), and sequential Phragmen's Rule. Weighted variants of these methods are also provided, allowing for differential voter influence.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://bblodfon.github.io/fastVoteR/}
}

}
\author{
\strong{Maintainer}: John Zobolas \email{bblodfon@gmail.com} (\href{https://orcid.org/0000-0002-3609-8674}{ORCID})

Other contributors:
\itemize{
  \item Anne-Marie George (\href{https://orcid.org/0000-0001-9232-8211}{ORCID}) [contributor]
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/voting_methods.R
\name{seq_phragmen}
\alias{seq_phragmen}
\title{Sequential Phragmen's Rule}
\usage{
seq_phragmen(
  voters,
  candidates,
  weights = NULL,
  committee_size = NULL,
  borda_score = TRUE,
  check = FALSE
)
}
\arguments{
\item{voters}{(\code{list()}) \cr
A list of subsets (\code{character} vectors), where each subset contains the
candidates approved or selected by a voter.}

\item{candidates}{(\code{character()}) \cr
A vector of all candidates to be ranked.}

\item{weights}{(\code{numeric()|NULL}) \cr
A numeric vector of non-negative weights representing each voter's influence.
Larger weight, higher influence.
Must have the same length as \code{voters}.
If \code{NULL} (default), all voters are assigned equal weights of 1, representing
equal influence.}

\item{committee_size}{(\code{integer(1)|NULL})\cr
Number of top-ranked candidates to return.
Default (\code{NULL}) returns all candidates.}

\item{borda_score}{(\code{logical(1)}) \cr
Whether to include a \code{borda_score} column in the output, which provides a
normalized score based on the candidate's rank.
If \code{TRUE} (default), the \code{borda_score} is calculated as \eqn{(p - i) / (p - 1)},
where \eqn{p} is the total number of candidates and \eqn{i} is the candidate's rank.}

\item{check}{(\code{logical(1)}) \cr
Whether to run additional voter-integrity checks. When \code{TRUE}, each voter
must approve at least one candidate, approvals must be unique per voter, and
all approved candidates must appear in \code{candidates}. Use \code{FALSE} to skip
these checks when inputs are known to be valid.}
}
\value{
A \code{data.frame} with columns:
\itemize{
\item \code{"candidate"}: Candidate names.
\item \code{"borda_score"}: Borda scores for method-agnostic comparison, ranging in \eqn{[0,1]}, where the top candidate receives a score of 1 and the lowest-ranked candidate receives a score of 0, based on the total number of candidates.
}

Candidates are ordered by the sequence in which they were selected.
}
\description{
\strong{Sequential Phragmen's Rule} is a multi-winner method that builds a committee
by distributing representation \emph{loads} across voters as evenly as possible.
At each step, it selects the candidate that yields the smallest increase in
voter load, then updates loads for voters who approved the chosen candidate.
The process continues until the committee is filled or all candidates are ranked.

This function uses an internal C++ implementation for efficient computation.
}
\seealso{
Other voting methods: 
\code{\link{av}()},
\code{\link{sav}()},
\code{\link{seq_pav}()}
}
\concept{voting methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rank_candidates.R
\name{rank_candidates}
\alias{rank_candidates}
\title{Rank candidates based on voter preferences}
\usage{
rank_candidates(
  voters,
  candidates,
  weights = NULL,
  committee_size = NULL,
  method = "av",
  borda_score = TRUE,
  shuffle_candidates = TRUE,
  check = FALSE
)
}
\arguments{
\item{voters}{(\code{list()}) \cr
A list of subsets (\code{character} vectors), where each subset contains the
candidates approved or selected by a voter.}

\item{candidates}{(\code{character()}) \cr
A vector of all candidates to be ranked.}

\item{weights}{(\code{numeric()|NULL}) \cr
A numeric vector of non-negative weights representing each voter's influence.
Larger weight, higher influence.
Must have the same length as \code{voters}.
If \code{NULL} (default), all voters are assigned equal weights of 1, representing
equal influence.}

\item{committee_size}{(\code{integer(1)|NULL})\cr
Number of top-ranked candidates to return.
Default (\code{NULL}) returns all candidates.}

\item{method}{(\code{character(1)}) \cr
The ranking voting method to use. Must be one of: \code{"av"}, \code{"sav"}, \code{"seq_pav"},
\code{"seq_phragmen"}. See Details.}

\item{borda_score}{(\code{logical(1)}) \cr
Whether to include a \code{borda_score} column in the output, which provides a
normalized score based on the candidate's rank.
If \code{TRUE} (default), the \code{borda_score} is calculated as \eqn{(p - i) / (p - 1)},
where \eqn{p} is the total number of candidates and \eqn{i} is the candidate's rank.}

\item{shuffle_candidates}{(\code{logical(1)}) \cr
Whether to randomly shuffle candidates before ranking.
This provides random tie-breaking and avoids deterministic bias when scores
are equal. Default is \code{TRUE}.}

\item{check}{(\code{logical(1)}) \cr
Whether to run additional voter-integrity checks. When \code{TRUE}, each voter
must approve at least one candidate, approvals must be unique per voter, and
all approved candidates must appear in \code{candidates}. Use \code{FALSE} to skip
these checks when inputs are known to be valid.}
}
\value{
A \code{data.frame} with columns:
\itemize{
\item \code{"candidate"}: Candidate names.
\item \code{"score"}: Scores assigned to each candidate based on the selected method
(if applicable).
\item \code{"norm_score"}: Normalized scores (if applicable), scaled to the range
\eqn{[0,1]}, which can be loosely interpreted as \strong{selection probabilities}
(see Meinshausen et al. (2010) for an example in Machine Learning research
where the goal is to perform stable feature selection).
\item \code{"borda_score"}: Borda scores for method-agnostic comparison, ranging in
\eqn{[0,1]}, where the top candidate receives a score of 1 and the lowest-ranked
candidate receives a score of 0.
}

Candidates are ordered by decreasing \code{"score"}, or by \code{"borda_score"} if the
method returns only rankings.
}
\description{
Calculates a ranking of candidates based on voters' preferences.
Approval-Based Committe (ABC) rules are based on Lackner et al. (2023).
For an example use in Machine Learning for ensemble feature selection,
see Zobolas et al. (2026).
}
\details{
We implement several consensus-based ranking methods, where voters express
preferences for candidates.
The framework has three components:
\itemize{
\item \strong{Voters}: A list where each element is the set of approved candidates for
a single voter.
\item \strong{Candidates}: A character vector of all possible candidates. This vector
can be shuffled to randomize tie-breaking across methods.
\item \strong{Weights}: A numeric vector giving each voter’s influence. Equal weights
mean equal influence; differing weights reflect varying importance.
}

This function is a thin wrapper that dispatches to method-specific implementations.
Supported methods include:
\enumerate{
\item Approval Voting (\code{method = "av"}), calls \code{\link[=av]{av()}}
\item Satisfaction Approval Voting (\code{method = "sav"}), calls \code{\link[=sav]{sav()}}
\item Sequential Proportional Approval Voting (\code{method = "seq_pav"}), calls \code{\link[=seq_pav]{seq_pav()}}
\item Sequential Phragmen’s Rule (\code{method = "seq_phragmen"}), calls \code{\link[=seq_phragmen]{seq_phragmen()}}
}

All methods support voter weights.

For method-agnostic comparisons, we compute \strong{borda scores}, which map each
candidate’s rank to a normalized scale across methods that may otherwise
return different score scales or only ordinal rankings.

For sequential methods such as \code{"seq_pav"} and \code{"seq_phragmen"}, the
\code{committee_size} parameter can speed up computation by selecting only the top
\eqn{N} candidates instead of producing a full ranking. For non-sequential methods
(e.g., \code{"sav"} or \code{"av"}), it simply truncates the final ranking to the top
\eqn{N} candidates.
}
\examples{
# 5 candidates
candidates = paste0("V", seq_len(5))

# 4 voters
voters = list(
  c("V3", "V1", "V4"),
  c("V3", "V1"),
  c("V3", "V2"),
  c("V2", "V4")
)

# voter weights
weights = c(1.1, 2.5, 0.8, 0.9)

# Approval voting (all voters equal)
rank_candidates(voters, candidates)

# Approval voting (voters unequal)
rank_candidates(voters, candidates, weights)

# Satisfaction Approval voting (voters unequal, no borda score)
rank_candidates(voters, candidates, weights, method = "sav", borda_score = FALSE)

# Sequential Proportional Approval voting (voters equal, no borda score)
rank_candidates(voters, candidates, method = "seq_pav", borda_score = FALSE)

# Sequential Phragmen's Rule (voters equal)
rank_candidates(voters, candidates, method = "seq_phragmen", borda_score = FALSE)

}
\references{
Lackner M, Skowron P (2023). \emph{Multi-Winner Voting with Approval Preferences}.
Springer Nature, 121 p. \doi{10.1007/978-3-031-09016-5}.

Zobolas J, George A, Lopez A, Fischer S, Becker M, Aittokallio T (2026).
"Prognostic biomarker discovery in pancreatic cancer through hybrid ensemble
feature selection and multi-omics data." \emph{BioData Mining}. \doi{10.1186/s13040-026-00546-0}.

Meinshausen N, Buhlmann P (2010). "Stability Selection." \emph{Journal of the Royal
Statistical Society Series B: Statistical Methodology}, 72(4), 417-473.
\doi{10.1111/J.1467-9868.2010.00740.X}.
}
